/*********************************************************************
* BCL.C - best cell classification method and first occurence
* method for population boundary estimation
* Estimates boundary in orthogonal (LINEAR) or polar (CIRCULAR)
* coordinates.
* Programmer: Alexei Sharov
* Organization: Dept. of Entomology, Virginia Tech. Blacksburg, VA 24060
* Tel.: (703) 231-7316
* E-mail sharov@vt.edu
LOG: 1/31/95: change_coord was changed to use rotation with a
   matrix. Also, another option to generate a map and fill bounded
   area was included. 
*********************************************************************/
#include <stdlib.h>
#include <dos.h>
#include "mylib.c"
#include "matrix.c"
#include "board.c"

#define LGE		  0.434294
#define MAXCUT   10
#define LINEAR   1
#define CIRCULAR 2
#define NODATA   -1.0e10
#define BC       1
#define FO       2

typedef struct param_st{
	char  datafile[40], outfile[40];
	char  consolfile[40], indicfile[40]; /* optional output maps */
	int   bc_map, indicmap;              /* indicators to build maps */
	double a, b;                         /* linear coeff. of gen. boundary */
	float x0,y0,x1,y1;                   /* gen boundary points */
	float xcent, ycent;                  /* circular boundary center */
	float output_width, strip_width, spline_width;
	int   bound_type;            /* LINEAR or CIRCULAR */
	int   sign;                  /* +1 if popul. is above the bound, else -1*/
	float tmin;                  /* trimming limit, should be low enough */
	float rel_weight[MAXCUT];    /* "island" misclassif. rel. weight */
	float precis;                /* pricision for regression analysis */
	float mindist;               /* min. dist from boundary to area border */
	float mult;                  /* multiplier for all distances */
	int   ncut;                  /* number of cutoffs */
	float cut[MAXCUT];           /* cut level */
	int   princ_comp;            /* index to adjust the gen. boundary */
	int   print_head;            /* print header */
	int   maxiter;               /* max iterations */
	int   anal_type;             /* BC (best class.) or FO (first occ.)*/
	float bndlen;                /* max bound. length for output */

  /* The rest are not parameters, but they are placed here for convenience:*/
	float da[MAXCUT];
	float db[MAXCUT];
	int   nstrip, iter;
	float *bound, act_prec;
	int   xmin, xmax;
}PARAM;

/**********************************/
double  atanfn  (float dx, float dy)
/**********************************/
/* Improved acrtangence function
*/
{
	double alpha;

	if (dx == 0 && dy == 0)
		printf("Error atanfn\n");
	if (dx == 0)
	{
		if (dy > 0) return(PI/2);
		else return(-PI/2);
	}
	if (dy == 0)
	{
		if (dx > 0) return(0);
		else return(PI);
	}
	alpha = atan(dy/dx);
	if (dx < 0 && dy > 0)
		return (alpha + PI);
	else if (dx < 0 && dy < 0)
		return (alpha - PI);
	else
		return (alpha);
}

/*********************************/
void grid2coord (int ix, int iy, float *x, float *y, DOC *d, int first)
/*********************************/
/* Converts grid cell indexex ix and iy into actual map coordinates
* x and y.
*/
{
	static double dx, dy;

	if (first)
	{
		dx = (d->xmax - d->xmin)/d->nx;
		dy = (d->ymax - d->ymin)/d->ny;
	}
	*x = (ix+0.5)*dx + d->xmin;
	*y = (iy+0.5)*dy + d->ymin;
}

/*********************************/
void change_coord (float x0, float y0, float *x, float *y, PARAM *p,
						int first)
/*********************************/
/* Changes coordinates from actual to rotated using the general
* boundary equasion: Y = a + bX, where p->a = a and p->b = b.
* Polar coordinates are: x = azimuth, y = distance from center.
*/
{
	static double cosa, sina;

	if (first)
	{
		cosa = 1.0 / sqrtfn(1.0 + p->b*p->b);
		sina = cosa * p->b;
	}
	if (p->bound_type == LINEAR)
	{
		x0 -= p->x0;
		y0 -= p->y0;
		*x = x0 * cosa + y0 * sina;
		*y = -x0 * sina + y0 * cosa;
	}
	else  /* p->bound_type == CIRCULAR */
	{
		x0 -= p->xcent;
		y0 -= p->ycent;
		*x = atanfn(x0,y0)*180/PI;
		*x = 90 - *x;  /* convert to azimuth */
		if (*x < 0)
			*x += 360;
		*y = (float)sqrt(x0*x0+y0*y0);
	}
}

/*********************************/
void projection (float x0, float y0, float *x, float *y, PARAM *p)
/*********************************/
/* Estimates coodrinates of point (x0,y0) projection on the
* general boundary: Y = a + bX, where p->a = a and p->b = b.
*/
{
	float sina, xx, yy;

	sina = p->b / sqrtfn(1.0 + p->b*p->b);
	change_coord(x0, y0, &xx, &yy, p, TRUE);
	*x = x0 + yy * sina;
	*y = p->y0 + p->b * (*x - p->x0);
}

/*********************************/
void new_coordinates (int ix, int iy, float *x, float *y, PARAM *p,
							BOARD *b, int first)
/*********************************/
/* Takes grid cell indexes and converts them into a new coordinate
* system. X-axis is expressed in strip width units, and is adjusted
* so that it can be used as an index of a strip.
*/
{
	float xx, yy, xc, yc, xxc, yyc;
	DOC *d;

	d = get_doc(b);
	grid2coord(ix, iy, &xx, &yy, d, first);
	change_coord(xx, yy, x, y, p, first);
	*x = *x / p->strip_width - p->xmin;
	*y /= p->mult;
}

/**************************************/
int  find_close (int ix, int iy, int *kx, int *ky, int maxdist)
/**************************************/
/* Spiral grid cell selection algorithm. Goes from the center (ix,iy).
* Returns 0 normally, and 1 if maximum distance is exceeded.
*/
{
	int dx, dy;

	dx = *kx - ix;
	dy = *ky - iy;
	if (dx != 0 || dy != 0)
	{
		if (dx >= dy && dx < -dy)
			++(*kx);
		else if (dx > dy && dx >= -dy)
			++(*ky);
		else if (dx <= dy && dx > -dy)
			--(*kx);
		else
			--(*ky);
	}
	if (dx==0 && dy==0 || dx==dy-1 && dx < 0)
	{
		(*kx)--;
		(*ky)--;
		if (ix-(*kx) >= maxdist)
			return(1);
	}
	return(0);
}

/*********************************/
int get_next_cell (int *ix, int *iy, DOC *d, PARAM *p, int first)
/*********************************/
/* Selects next grid cell.
*/
{
	long ixcent, iycent;
	int done = FALSE;
	static float dx, dy, maxdist, xc, yc;

	if (first)
	{
		dx = (d->xmax - d->xmin)/d->nx;
		dy = (d->ymax - d->ymin)/d->ny;
		if (p->bound_type == LINEAR)
		{
			*ix = 0;
			*iy = 0;
			return(done);
		}
		else
		{
			*ix = (p->xcent - d->xmin)/dx;
			*iy = (p->ycent - d->ymin)/dy;
			xc = *ix;
			yc = *iy;
			maxdist = abs(xc);
			if (abs(yc) > maxdist)
				maxdist = abs(yc);
			if (abs(xc - d->nx-1) > maxdist)
				maxdist = abs(xc - d->nx-1);
			if (abs(yc - d->ny-1) > maxdist)
				maxdist = abs(yc - d->ny-1);
		}
	}
	if (p->bound_type == CIRCULAR)
	{
		if (!first)
			done = find_close(xc, yc, ix, iy, maxdist);
		while (!done && !correct_entry(*ix, *iy, d))
			done = find_close(xc, yc, ix, iy, maxdist);
	}
	else
	{
		if (absfn(p->b) < 1)
		{
			(*ix)++;
			if (*ix == d->nx)
			{
				*ix = 0;
				(*iy)++;
			}
			if (*iy == d->ny)
				done = TRUE;
		}
		else
		{
			(*iy)++;
			if (*iy == d->ny)
			{
				*iy = 0;
				(*ix)++;
			}
			if (*ix == d->nx)
				done = TRUE;
		}
	}
	return(done);
}

/*********************************/
void  best_classification (BOARD *b, PARAM *p)
/*********************************/
{
	int ix, iy, i, ic, first = TRUE, ii, ixx;
	float val, x, y, dist, *minimum, x1, y1, criterion;
	int *absent;   /* number of non-occupied cells in a strip */
	int *acc_abs;  /* accumulated number of non-occupied cells */
	int *acc_pres; /* accumulated number of occupied cells */
	FILE *fp;
	DOC *d;

	d = get_doc(b);
	/* Initialize arrays */
	check(absent =  (int*)calloc(p->nstrip*p->ncut, sizeof(int)));
	check(acc_abs = (int*)calloc(p->nstrip*p->ncut, sizeof(int)));
	check(acc_pres =(int*)calloc(p->nstrip*p->ncut, sizeof(int)));
	check(minimum = (float*)calloc(p->nstrip*p->ncut, sizeof(float)));

	/* Accumulate non-occupied cell counts */
	if (p->anal_type == BC)
	{
		for(ix = 0; ix<d->nx; ++ix)
		{
			for(iy = 0; iy<d->ny; ++iy)
			{
				val = get_cell(ix,iy,b);
				if (val < p->tmin)
					continue;
				new_coordinates (ix, iy, &x, &y, p, b, first);
				ixx = (int)x;
				first = FALSE;
				if (ixx < 0 || ixx >= p->nstrip)
					continue;
				for (ic=0; ic<p->ncut; ++ic)
				{
					ii = ixx*p->ncut + ic;
					if (val < p->cut[ic])
						absent[ii]++;
				}
			}
		}
	}

	/* Initialize boundary and "minimum" arrays */
	for(ix = 0; ix < p->nstrip; ++ix)
	{
		for (ic=0; ic<p->ncut; ++ic)
		{
			ii = ix*p->ncut + ic;
			p->bound[ii] = NODATA;
			minimum[ii] = 1.0e30;
		}
	}

	/* Find the boundary */
	first = TRUE;
	if (p->anal_type == FO)
		new_coordinates (0, 0, &x, &y, p, b, first);
	/* Go through all the cells: */
	while(!get_next_cell(&ix, &iy, d, p, first))
	{
		first = FALSE;
		val = get_cell(ix,iy,b);
		if (val < p->tmin)
			continue;
		new_coordinates (ix, iy, &x, &y, p, b, first);
		ixx = (int)x;
		if (ixx < 0 || ixx >= p->nstrip)
			continue;
		for (ic=0; ic<p->ncut; ++ic)
		{
			ii = ixx*p->ncut + ic;
			if (p->anal_type == BC)
			{
				/* Accumulate occupied and non-occupied cells */
				if (val >= p->cut[ic])
					acc_pres[ii]++;
				else
					acc_abs[ii]++;

				/* Estimate the weighted number of misclassified cells */
				criterion = p->rel_weight[ic]*acc_pres[ii] +
				 (absent[ii] - acc_abs[ii]);
				/* Change the sign of "criterion" if cells are accumulated
				* starting from generally occupied area. */
				if (p->bound_type == CIRCULAR && p->sign > 0)
					criterion = -criterion;
				if (p->bound_type == LINEAR)
				{
					if (p->b > 1)
						criterion *= -p->sign;
					else
						criterion *= p->sign;
				}
			}
			else  /* FO first occurence */
			{
				if (val >= p->cut[ic])
				{
					if (p->bound_type == CIRCULAR)
						criterion = -y * p->sign;
					else  /* if p->bound_type == LINEAR */
						criterion = y * p->sign;
				}
				else
					criterion = 1.0e30;
			}

			if (criterion < minimum[ii])
			{
				p->bound[ii] = y;
				minimum[ii] = criterion;

				/* Check if boundary points are greater than tmin. Otherwise
				* they can be confused with trimming minimum */
				if (p->bound[ii] < p->tmin)
				{
					printf("Boundary values are less than tmin!\n");
					exit(0);
				}
			}
		}
	}

	/* Check if boundary is too close to the edge of the area */
	for(iy = -1; iy<=d->ny; ++iy)
	{
		for(ix = -1; ix<=d->nx; ++ix)
		{
			if (iy > -1 && iy < d->ny && ix > -1 && ix < d->nx)
				val = get_cell(ix,iy,b);
			else
				val = p->tmin - 1;
			if (val < p->tmin)
			{
				new_coordinates (ix, iy, &x, &y, p, b, first);
				ixx = (int)x;
				if (ixx < 0 || ixx >= p->nstrip)
					continue;
				for (ic=0; ic<p->ncut; ++ic)
				{
					ii = ixx*p->ncut + ic;
					if (ii < 0 || ii >= p->nstrip*p->ncut)
						continue;
					if (absfn(p->bound[ii] - y) < p->mindist/p->mult)
						p->bound[ii] = NODATA;
				}
			}
		}
	}
	free(acc_abs);
	free(acc_pres);
	free(absent);
	free(minimum);
}

/*********************************/
void  make_maps (BOARD *b, PARAM *p)
/*********************************/
{
	int ix, iy, i, ic, ixx, first = TRUE, ii, sign;
	float valind, valcons;
	float delta;
	float val, x, y;
	FILE *fcons, *find;
	DOC doc, *d;
	char docfile[40];

	d = get_doc(b);
	if (p->bc_map)
		fcons = open_board_file (p->consolfile, d, "w");
	if (p->indicmap)
		find = open_board_file (p->indicfile, d, "w");
	sign = p->sign;
	if (p->bound_type == CIRCULAR)
		sign *= -1;
	for(iy = d->ny-1; iy >=0; --iy)
	{
		for(ix = 0; ix<d->nx; ++ix)
		{
			new_coordinates (ix, iy, &x, &y, p, b, first);
			val = get_cell(ix,iy,b);
			if (val < p->tmin)
			{
				valind = val;
				valcons = val;
			}
			else
			{
				ic = 0;
				while (ic < p->ncut && val >= p->cut[ic])
					++ic;
				valind = ic;
				ixx = (int)x;
				if (ixx < 0)
					ixx = 0;
				if (ixx >= p->nstrip)
					ixx = p->nstrip-1;
				ic = 0;
				if (sign > 0)
					while (ic < p->ncut && y > (p->bound[ixx*p->ncut + ic]))
						++ic;
				else
					while (ic < p->ncut && y < (p->bound[ixx*p->ncut + ic]))
						++ic;
				valcons = ic;
			}

			/* put a line */
			delta = 0.5;
			if (y <= delta && y > -delta)
			{
				if (p->bc_map >= 2)
					valcons = p->ncut+1;
				if (p->indicmap >= 2)
					valind = p->ncut+1;
			}
			if (p->bc_map)
				fwrite(&valcons, sizeof(float), 1, fcons);
			if (p->indicmap)
				fwrite(&valind, sizeof(float), 1, find);
		}
	}
	doc = *d;
	doc.data_type = REAL;
	doc.file_type = BINARY;
	if (p->bc_map)
	{
		fclose (fcons);
		modify_name (p->consolfile, docfile, ".doc");
		print_doc (&doc, docfile);
	}
	if (p->indicmap)
	{
		fclose (find);
		modify_name (p->indicfile, docfile, ".doc");
		print_doc (&doc, docfile);
	}
}

/*********************************/
int  pricip_compon (PARAM *p)
/*********************************/
{
	int ix, ic, ii, i, num;
	float *xx, *yy, cov, b0, b, y, rsq;
	float avera, averb, mult = 0.67, x0, y0;
	static float bold;

	check(xx = (float*)calloc(p->nstrip, sizeof(float)));
	check(yy = (float*)calloc(p->nstrip, sizeof(float)));
	change_coord(p->x0, p->y0, &x0, &y0, p, TRUE);
	for (ic=0; ic<p->ncut; ++ic)
	{
		num = 0;
		for(ix = 0; ix < p->nstrip; ++ix)
		{
			ii = ix*p->ncut + ic;
			if (p->bound[ii] == NODATA)
				continue;
			y = p->bound[ii]*p->mult;
			xx[num] = (ix + p->xmin + 0.5)*p->strip_width + x0;
			yy[num] = y;
			++num;
		}
		if (num > 4)
		{
			matrRegress(xx, yy, &b, &b0, &cov, &rsq, 1, num);
			p->db[ic] = b;
		}
		else
		{
			p->db[ic] = NODATA;
		}
	}
	free(xx);
	free(yy);

	/* Update linear regression */
	averb = 0;
	num = 0;
	for (ic=0; ic<p->ncut; ++ic)
	{
		if (p->db[ic] == NODATA)
			continue;
		averb += p->db[ic];
		++num;
	}
	if (num <= 0)
	{
		printf ("Unable to get the boundary!\n");
		exit(0);
	}
	averb *= mult/num;
	p->b = (p->b + averb)/(1.0 - p->b * averb);
	projection (p->x1, p->y1, &x0, &y0, p);
	p->x1 = x0;
	p->y1 = y0;
	p->act_prec = averb/mult;
	if (absfn(p->act_prec) < p->precis)
		return(TRUE);
	if (absfn(p->act_prec - bold) < p->precis/2)
		return(TRUE);
	bold = p->act_prec;
	return(FALSE);
}

/*********************************/
void  spline  (PARAM *p)
/*********************************/
{
	float *bnd, bnd1, delta;
	int num, ii, ic, ix, ixx;

	check(bnd = (float*)calloc(p->nstrip, sizeof(float)));
	delta = p->spline_width/p->strip_width/2;
	for (ic=0; ic<p->ncut; ++ic)
	{
		for(ix = 0; ix < p->nstrip; ++ix)
		{
			bnd1 = 0;
			num = 0;
			for(ixx = ix-delta; ixx < ix+delta; ++ixx)
			{
				if (ixx < 0 || ixx >= p->nstrip)
					continue;
				ii = ixx*p->ncut+ic;
				if (p->bound[ii] == NODATA)
					continue;
				bnd1 += p->bound[ii];
				num++;
			}
			if (num > 0)
				bnd[ix] = bnd1/num;
			else
				bnd[ix] = NODATA;
		}
		for(ix = 0; ix < p->nstrip; ++ix)
		{
			ii = ix*p->ncut+ic;
			p->bound[ii] = bnd[ix];
		}
	}
	free (bnd);
}

/*********************************/
void  extrapolate  (PARAM *p)
/*********************************/
{
	int delta;
	int ix, ii, ic, ixx, ix1, ix2, iwid;
	float y1, y2, y3;

	iwid = p->output_width/p->strip_width + 0.5;
	for (ic=0; ic<p->ncut; ++ic)
	{
		for(ix = 0; ix < p->nstrip; ++ix)
		{
			delta = 0;
			while (delta < p->nstrip && p->bound[ii=ix*p->ncut+ic]==NODATA)
			{
				if (delta <= 0)
					delta = -1*delta + 1;
				else
					delta *= -1;
				ixx = ix+delta;
				if (ixx < 0 || ixx >= p->nstrip || p->bound[ixx*p->ncut+ic]==NODATA)
					continue;
				p->bound[ii] = p->bound[ixx*p->ncut+ic];
			}
		}
		for(ix = 0; ix < p->nstrip; ++ix)
		{
			if ((ix + p->xmin + iwid/2 + 1)%iwid == 0)
				continue;
			ix1 = ((ix + p->xmin + iwid/2 + 1)/iwid)*iwid - p->xmin - iwid/2 - 1;
			ix2 = ix1+iwid;
			if (ix1 <0 && ix2 <0 || ix1 >=p->nstrip && ix1 >=p->nstrip)
				continue;
			if (ix1 <0) ix1 = 0;
			if (ix2 >=p->nstrip) ix2=p->nstrip-1;
			y1 = p->bound[ix1*p->ncut+ic];
			y2 = p->bound[ix2*p->ncut+ic];
			if (ix1==ix2)
				continue;
			y3 = y1 + (y2-y1)/(ix2-ix1)*(float)(ix-ix1);
			p->bound[ix*p->ncut+ic] = y3;
		}
	}
}

/*********************************/
void  print_output  (PARAM *p)
/*********************************/
{
	FILE *fp;
	int ix, ic, ii;
	float boundary, dx, dy, x;
	int iwid, printout;

	fp = fopen(p->outfile, "w");
	if (p->print_head)
	{
		if (p->princ_comp)
		{
			fprintf(fp,"Iterations = %d\n", p->iter);
			fprintf(fp,"Precision = %f\n", p->act_prec);
		}
		if (p->bound_type == LINEAR)
		{
			fprintf(fp,"Slope:  b = %f\n", (float)p->b);
			fprintf(fp,"Bound.points (x1,y1; x2,y2): %10.0f,%10.0f;  %10.0f,%10.0f\n",
			  p->x0, p->y0, p->x1, p->y1);
		}
		else
			fprintf(fp,"Center (x,y): %f,%f\n", p->xcent, p->ycent);
		fprintf(fp,"Cut levels: ");
		for (ic=0; ic<p->ncut; ++ic)
			fprintf(fp, "%9.3f ", p->cut[ic]);
		fprintf(fp,"\n");
	}

	iwid = p->output_width/p->strip_width + 0.5;
	for(ix = 0; ix < p->nstrip; ++ix)
	{
		if ((ix + p->xmin + iwid/2 + 1)%iwid != 0)
			continue;
		x = (ix + p->xmin + 0.5*(iwid%2))*p->strip_width;
		if (p->bound_type == LINEAR)
			x /= p->mult;
		fprintf(fp,"%10.2f  ", x);
		for (ic=0; ic<p->ncut; ++ic)
		{
			ii = ix*p->ncut+ic;
			boundary = p->bound[ii];
			if (boundary == NODATA)
				boundary = p->tmin-1;
			fprintf(fp,"%9.3f ", boundary);
		}
		fprintf(fp,"\n");
	}
	fclose(fp);
}

/*********************************/
void  get_xminmax (BOARD *b, PARAM *p)
/*********************************/
{
	float x, y, xx, yy, azmin = 400, azmax = -400;
	int ix, iy, first=TRUE;
	DOC *d;

	d = get_doc(b);
	if (p->bound_type == LINEAR)
	{
		grid2coord(0, 0, &xx, &yy, d, first);
		change_coord(xx, yy, &x, &y, p, first);
		p->xmin = (int)floor(x/p->strip_width);
		grid2coord(0, d->ny-1, &xx, &yy, d, first);
		change_coord(xx, yy, &x, &y, p, first);
		if (x/p->strip_width < p->xmin)
			p->xmin = (int)floor(x/p->strip_width);
		p->nstrip = p->bndlen/p->strip_width + 0.5;
	}
	else /* if p->bound_type == LINEAR */
	{
		if (p->xcent >= d->xmin && p->xcent < d->xmax && p->ycent < d->ymax)
		{
			p->xmin = 0;
			p->nstrip = 360/p->strip_width;
		}
		else
		{
			for(ix=0; ix<d->nx; ix += d->nx-1)
			{
				for(iy=0; iy<d->ny; iy += d->ny-1)
				{
					grid2coord(ix, iy, &xx, &yy, d, first);
					change_coord(xx, yy, &x, &y, p, first);
					first = FALSE;
					if (azmin > x)
						azmin = x;
					if (azmax < x)
						azmax = x;
				}
			}
			p->xmin = azmin/p->strip_width;
			p->nstrip = azmax/p->strip_width - p->xmin + 1;
		}
	}
	p->xmax = p->xmin + p->nstrip;
}

/*********************************/
PARAM *read_parameters (char *filename)
/*********************************/
{
	PARAM *p;
	FILE *fp;
	int ic;

	fp = fopen(filename,"r");
	if (!fp)
	{
		printf("Bad file name (param)\n");
		exit(0);
	}
	check(p=(PARAM*)calloc(1,sizeof(PARAM)));
	read_string (p->consolfile, 40, fp);
	read_string (p->indicfile, 40, fp);
	p->bc_map = read_int(fp);
	p->indicmap = read_int(fp);
	p->tmin = read_float(fp);
	p->bound_type = read_int(fp);
	p->anal_type = read_int(fp);
	p->x0 = read_float(fp);
	p->y0 = read_float(fp);
	p->x1 = read_float(fp);
	p->y1 = read_float(fp);
	p->xcent = read_float(fp);
	p->ycent = read_float(fp);
	p->strip_width = read_float(fp);
	p->output_width = read_float(fp);
	p->spline_width = read_float(fp);
	p->sign = read_int(fp);
	p->mult = read_float(fp);
	p->bndlen = read_float(fp);
	p->precis = read_float(fp);
	p->maxiter = read_int(fp);
	p->ncut = read_int(fp);
	if (p->ncut > MAXCUT)
	{
		printf("ncut is too large!\n");
		exit(0);
	}
	for(ic=0; ic < p->ncut; ++ic)
	{
		p->cut[ic] = read_float(fp);
		p->rel_weight[ic] = read_float(fp);
	}
	p->princ_comp = read_int(fp);
	p->print_head = read_int(fp);
	p->mindist = read_float(fp);
	fclose(fp);
	return(p);
}

/*********************************/
int  main   (int argc, char *argv[])
/*********************************/
{
	BOARD *b;
	PARAM *p;
	char parfile[40] = {"bcl.par"}, *ps;
	int done;

	if (argc < 3)
	{
		printf("bcl infile outfile [parfile]\n");
		exit(0);
	}
	if (argc >= 4)
		strcpy(parfile, argv[3]);
	p = read_parameters (parfile);
	strcpy(p->datafile, argv[1]);
	strcpy(p->outfile, argv[2]);

	b = new_board(p->datafile);
	p->b = (p->y1 - p->y0)/(p->x1 - p->x0);

	get_xminmax(b, p);
	check(p->bound = (float*)malloc(p->nstrip*p->ncut*sizeof(float)));
	if (!p->princ_comp || p->bound_type == CIRCULAR)
		best_classification (b, p);
	else
	{
		done = FALSE;
		p->iter = 0;
		while (!done && p->iter <= p->maxiter)
		{
			best_classification (b, p);
			done = pricip_compon (p);
			++p->iter;
		}
	}
	spline(p);
	print_output(p);
	if (p->bc_map || p->indicmap)
	{
		extrapolate (p);
		make_maps (b, p);
	}
	return(0);
}

